import { UsageFlags, getCrossLanguagePackageId, isAzureCoreModel, } from "@azure-tools/typespec-client-generator-core";
import { ignoreDiagnostics } from "@typespec/compiler";
import { emitBasicHttpMethod, emitLroHttpMethod, emitLroPagingHttpMethod, emitPagingHttpMethod, } from "./http.js";
import { KnownTypes, disableGenerationMap, emitEndpointType, getType, simpleTypesMap, typesMap, } from "./types.js";
import { emitParamBase, getImplementation, removeUnderscoresFromNamespace } from "./utils.js";
function emitBasicMethod(context, rootClient, method, operationGroupName) {
    if (method.operation.kind !== "http")
        throw new Error("We only support HTTP operations right now");
    switch (method.operation.kind) {
        case "http":
            return emitBasicHttpMethod(context, rootClient, method, operationGroupName);
        default:
            throw new Error("We only support HTTP operations right now");
    }
}
function emitLroMethod(context, rootClient, method, operationGroupName) {
    if (method.operation.kind !== "http")
        throw new Error("We only support HTTP operations right now");
    switch (method.operation.kind) {
        case "http":
            return emitLroHttpMethod(context, rootClient, method, operationGroupName);
        default:
            throw new Error("We only support HTTP operations right now");
    }
}
function emitPagingMethod(context, rootClient, method, operationGroupName) {
    if (method.operation.kind !== "http")
        throw new Error("We only support HTTP operations right now");
    switch (method.operation.kind) {
        case "http":
            return emitPagingHttpMethod(context, rootClient, method, operationGroupName);
        default:
            throw new Error("We only support HTTP operations right now");
    }
}
function emitLroPagingMethod(context, rootClient, method, operationGroupName) {
    if (method.operation.kind !== "http")
        throw new Error("We only support HTTP operations right now");
    switch (method.operation.kind) {
        case "http":
            return emitLroPagingHttpMethod(context, rootClient, method, operationGroupName);
        default:
            throw new Error("We only support HTTP operations right now");
    }
}
function emitMethodParameter(context, parameter) {
    if (parameter.kind === "endpoint") {
        if (parameter.type.kind === "union") {
            for (const endpointVal of parameter.type.variantTypes) {
                return emitEndpointType(context, endpointVal);
            }
        }
        else {
            return emitEndpointType(context, parameter.type);
        }
    }
    const base = {
        ...emitParamBase(context, parameter),
        implementation: getImplementation(context, parameter),
        clientDefaultValue: parameter.clientDefaultValue,
        location: parameter.kind,
    };
    if (parameter.isApiVersionParam) {
        return [
            {
                ...base,
                location: "query",
                wireName: "api-version",
                in_docstring: false,
            },
        ];
    }
    return [base];
}
function emitMethod(context, rootClient, method, operationGroupName) {
    switch (method.kind) {
        case "basic":
            return emitBasicMethod(context, rootClient, method, operationGroupName);
        case "lro":
            return emitLroMethod(context, rootClient, method, operationGroupName);
        case "paging":
            return emitPagingMethod(context, rootClient, method, operationGroupName);
        default:
            return emitLroPagingMethod(context, rootClient, method, operationGroupName);
    }
}
function emitOperationGroups(context, client, rootClient, prefix) {
    const operationGroups = [];
    for (const method of client.methods) {
        if (method.kind === "clientaccessor") {
            const operationGroup = method.response;
            const name = `${prefix}${operationGroup.name}`;
            let operations = [];
            for (const method of operationGroup.methods) {
                if (method.kind === "clientaccessor")
                    continue;
                operations = operations.concat(emitMethod(context, rootClient, method, name));
            }
            operationGroups.push({
                name: name,
                className: name,
                propertyName: operationGroup.name,
                operations: operations,
                operationGroups: emitOperationGroups(context, operationGroup, rootClient, name),
            });
        }
    }
    // root client should deal with mixin operation group
    if (prefix === "") {
        let operations = [];
        for (const method of client.methods) {
            if (method.kind === "clientaccessor")
                continue;
            operations = operations.concat(emitMethod(context, rootClient, method, ""));
        }
        if (operations.length > 0) {
            operationGroups.push({
                name: "",
                className: "",
                propertyName: "",
                operations: operations,
            });
        }
    }
    return operationGroups.length > 0 ? operationGroups : undefined;
}
function emitClient(context, client) {
    if (client.initialization) {
        context.__endpointPathParameters = [];
    }
    const parameters = client.initialization?.properties
        .map((x) => emitMethodParameter(context, x))
        .reduce((a, b) => [...a, ...b]) ?? [];
    const endpointParameter = client.initialization?.properties.find((x) => x.kind === "endpoint");
    const operationGroups = emitOperationGroups(context, client, client, "");
    let url;
    if (endpointParameter?.type.kind === "union") {
        url = endpointParameter.type.variantTypes[0].serverUrl;
    }
    else {
        url = endpointParameter?.type.serverUrl;
    }
    return {
        name: client.name,
        description: (client.summary ? client.summary : client.doc) ?? "",
        parameters,
        operationGroups,
        url,
        apiVersions: client.apiVersions,
        arm: context.arm,
    };
}
export function emitCodeModel(sdkContext) {
    // Get types
    const sdkPackage = sdkContext.sdkPackage;
    const codeModel = {
        namespace: removeUnderscoresFromNamespace(sdkPackage.rootNamespace).toLowerCase(),
        clients: [],
        subnamespaceToClients: {},
    };
    for (const client of sdkPackage.clients) {
        codeModel["clients"].push(emitClient(sdkContext, client));
        if (client.nameSpace === sdkPackage.rootNamespace) {
        }
        else {
            codeModel["subnamespaceToClients"][client.nameSpace] = emitClient(sdkContext, client);
        }
    }
    // loop through models and enums since there may be some orphaned models needs to be generated
    for (const model of sdkPackage.models) {
        // filter out spread models
        if (model.name === "" ||
            ((model.usage & UsageFlags.Spread) > 0 &&
                (model.usage & UsageFlags.Input) === 0 &&
                (model.usage & UsageFlags.Output) === 0)) {
            continue;
        }
        // filter out specific models not used in python, e.g., pageable models
        if (disableGenerationMap.has(model)) {
            continue;
        }
        // filter out core models
        if (isAzureCoreModel(model)) {
            continue;
        }
        getType(sdkContext, model);
    }
    for (const sdkEnum of sdkPackage.enums) {
        // filter out api version enum since python do not generate it
        if (sdkEnum.usage === UsageFlags.ApiVersionEnum) {
            continue;
        }
        // filter out core enums
        if (isAzureCoreModel(sdkEnum)) {
            continue;
        }
        getType(sdkContext, sdkEnum);
    }
    codeModel["types"] = [
        ...typesMap.values(),
        ...Object.values(KnownTypes),
        ...simpleTypesMap.values(),
    ];
    codeModel["crossLanguagePackageId"] = ignoreDiagnostics(getCrossLanguagePackageId(sdkContext));
    return codeModel;
}
//# sourceMappingURL=code-model.js.map